# ptufile/setup.py

"""Ptufile package Setuptools script."""

import os
import re
import sys

import numpy
from setuptools import Extension, setup

buildnumber = ''

DEBUG = bool(os.environ.get('PTUFILE_DEBUG', False))


def search(pattern, string, flags=0):
    """Return first match of pattern in string."""
    match = re.search(pattern, string, flags)
    if match is None:
        raise ValueError(f'{pattern!r} not found')
    return match.groups()[0]


def fix_docstring_examples(docstring):
    """Return docstring with examples fixed for GitHub."""
    start = True
    indent = False
    lines = ['..', '  This file is generated by setup.py', '']
    for line in docstring.splitlines():
        if not line.strip():
            start = True
            indent = False
        if line.startswith('>>> '):
            indent = True
            if start:
                lines.extend(['.. code-block:: python', ''])
                start = False
        lines.append(('    ' if indent else '') + line)
    return '\n'.join(lines)


with open('ptufile/ptufile.py', encoding='utf-8') as fh:
    code = fh.read()

version = search(r"__version__ = '(.*?)'", code).replace('.x.x', '.dev0')
version += ('.' + buildnumber) if buildnumber else ''

description = search(r'"""(.*)\.(?:\r\n|\r|\n)', code)

readme = search(
    r'(?:\r\n|\r|\n){2}"""(.*)"""(?:\r\n|\r|\n){2}from __future__',
    code,
    re.MULTILINE | re.DOTALL,
)
readme = '\n'.join(
    [description, '=' * len(description)] + readme.splitlines()[1:]
)

if 'sdist' in sys.argv:
    # update README, LICENSE, and CHANGES files

    with open('README.rst', 'w', encoding='utf-8') as fh:
        fh.write(fix_docstring_examples(readme))

    license = search(
        r'(# Copyright.*?(?:\r\n|\r|\n))(?:\r\n|\r|\n)+""',
        code,
        re.MULTILINE | re.DOTALL,
    )
    license = license.replace('# ', '').replace('#', '')

    with open('LICENSE', 'w', encoding='utf-8') as fh:
        fh.write('BSD 3-Clause License\n\n')
        fh.write(license)

    revisions = search(
        r'(?:\r\n|\r|\n){2}(Revisions.*)- …',
        readme,
        re.MULTILINE | re.DOTALL,
    ).strip()

    with open('CHANGES.rst', encoding='utf-8') as fh:
        old = fh.read()

    old = old.split(revisions.splitlines()[-1])[-1]
    with open('CHANGES.rst', 'w', encoding='utf-8') as fh:
        fh.write(revisions.strip())
        fh.write(old)

ext_modules = [
    Extension(
        'ptufile._ptufile',
        ['ptufile/_ptufile.pyx'],
        define_macros=[
            # ('CYTHON_TRACE_NOGIL', '1'),
            # ('CYTHON_LIMITED_API', '1'),
            # ('Py_LIMITED_API', '1'),
            ('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION'),
        ],
        extra_compile_args=['/Zi', '/Od'] if DEBUG else [],
        extra_link_args=['-debug:full'] if DEBUG else [],
        include_dirs=[numpy.get_include()],
    )
]

setup(
    name='ptufile',
    version=version,
    license='BSD',
    description=description,
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Christoph Gohlke',
    author_email='cgohlke@cgohlke.com',
    url='https://www.cgohlke.com',
    project_urls={
        'Bug Tracker': 'https://github.com/cgohlke/ptufile/issues',
        'Source Code': 'https://github.com/cgohlke/ptufile',
        # 'Documentation': 'https://',
    },
    packages=['ptufile'],
    package_data={'ptufile': ['py.typed']},
    entry_points={'console_scripts': ['ptufile = ptufile.__main__:main']},
    python_requires='>=3.10',
    install_requires=['numpy'],
    setup_requires=['setuptools', 'numpy'],
    extras_require={'all': ['xarray', 'tifffile', 'matplotlib']},
    ext_modules=ext_modules,
    zip_safe=False,
    platforms=['any'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
)



# ptufile/numcodecs.py

# Copyright (c) 2024, Christoph Gohlke
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""PTU codec for the Numcodecs package."""

from __future__ import annotations

__all__ = ['register_codec', 'Ptu']

from io import BytesIO
from typing import TYPE_CHECKING

from numcodecs import registry
from numcodecs.abc import Codec

from .ptufile import PtuFile

if TYPE_CHECKING:
    from collections.abc import Sequence
    from types import EllipsisType
    from typing import Any

    from numpy.typing import ArrayLike, DTypeLike, NDArray


class Ptu(Codec):  # type: ignore[misc]
    """Ptu codec for Numcodecs."""

    codec_id = 'ptufile'

    def __init__(
        self,
        *,
        selection: Sequence[int | slice | EllipsisType | None] | None = None,
        dtype: DTypeLike | None = None,
        channel: int | None = None,
        frame: int | None = None,
        dtime: int | None = None,
        trimdims: str | None = None,
        keepdims: bool = True,
    ):
        if selection is not None:
            # TODO: serialize slices, EllipsisType
            raise NotImplementedError(f'{selection=}')
        self.selection = selection
        self.dtype = dtype
        self.channel = channel
        self.frame = frame
        self.dtime = dtime
        self.trimdims = trimdims
        self.keepdims = bool(keepdims)

    def encode(self, buf: ArrayLike) -> None:
        """Return Ptu file as bytes."""
        raise NotImplementedError

    def decode(self, buf: bytes, out: Any | None = None) -> NDArray[Any]:
        """Return decoded image as NumPy array."""
        with BytesIO(buf) as fh:
            with PtuFile(fh) as ptu:
                result = ptu.decode_image(
                    self.selection,
                    dtype=self.dtype,
                    channel=self.channel,
                    frame=self.frame,
                    dtime=self.dtime,
                    keepdims=self.keepdims,
                )
        return result


def register_codec(cls: Codec = Ptu, codec_id: str | None = None) -> None:
    """Register :py:class:`Ptu` codec with Numcodecs."""
    registry.register_codec(cls, codec_id=codec_id)
